@@ -18,6 +18,7 @@ GNU Affero General Public License for more details.
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
+from pathlib import Path
 import time
 from typing import Dict, List, Union
 import requests
@@ -28,11 +29,12 @@ import tqdm
 import pandas
 from collections import defaultdict
 from datetime import datetime, timezone, date, timedelta
+from json.decoder import JSONDecodeError
 
 import plenty_api.keyring
 import plenty_api.utils as utils
 from plenty_api.constants import (
-    IMPORT_ORDER_DATE_TYPES, ORDER_TYPES, VALID_LANGUAGES
+    IMPORT_ORDER_DATE_TYPES, ORDER_TYPES, VALID_LANGUAGES, DUMPABLE_CONTENT_TYPES
 )
 
 
@@ -85,6 +87,12 @@ class PlentyApi():
 
         **plenty_api_get_shipping_packages_for_order**
 
+        **plenty_api_dump_bi_raw_file**
+
+        **plenty_api_get_bi_raw_files**
+
+        **plenty_api_get_amazon_product_types**
+
         POST REQUESTS
         **plenty_api_set_image_availability**
 
@@ -247,11 +255,17 @@ class PlentyApi():
         response = requests.post(endpoint, params=creds)
         if response.status_code == 403:
             logging.error(
-                "ERROR: Login to API failed: your account is locked\n"
+                "Login to API failed: your account is locked\n"
                 "unlock @ Setup->settings->accounts->{user}->unlock login"
             )
         try:
             token = utils.build_login_token(response_json=response.json())
+        except JSONDecodeError:
+            logging.error(
+                "Plentymarkets API refuses to grant a login token."
+                "(Possible reason: Too many users logged in)"
+            )
+            return False
         except KeyError:
             try:
                 if (
@@ -331,6 +345,11 @@ class PlentyApi():
             time.sleep(3)
 
         logging.debug(f"request url: {raw_response.request.url}")
+
+        # if the response is a file, return raw content, so it can be written to a file
+        if raw_response.headers['Content-Type'] in DUMPABLE_CONTENT_TYPES:
+            return raw_response.content
+
         try:
             response = raw_response.json()
         except simplejson.errors.JSONDecodeError:
@@ -371,17 +390,41 @@ class PlentyApi():
                 isinstance(response, list)):
             return response
 
-        page_info = utils.sniff_response_format(response=response)
-        entries = response[page_info['data']]
+        page_info = utils.sniff_response_format(response=response, query=query)
+        # Handle page slices (custom selection of pages)
+        slice_start = ''
+        slice_end = ''
+        if 'pages' in query:
+            pages = query['pages']
+            slice_start = pages['start_page'] if 'start_page' in pages else ''
+            slice_end = pages['end_page'] if 'end_page' in pages else ''
+            logging.debug(f"Using page slice [{slice_start}:{slice_end}]")
+        if slice_start and slice_start >= 2:
+            entries = []
+        else:
+            entries = response[page_info['data']]
 
-        if self.cli_progress_bar:
+        if self.cli_progress_bar and page_info['last_page']:
             pbar = None
             if not page_info['end_condition'](response):
                 pbar = tqdm.tqdm(desc=f'Plentymarkets {domain} request',
                                  total=response[page_info['last_page']])
+        elif self.cli_progress_bar and not page_info['last_page']:
+            logging.warn(f"Response for {domain} has no pagination, unable to detect the number of pages.")
 
+        page = 1
         while not page_info['end_condition'](response):
-            query.update({'page': response[page_info['page']] + 1})
+            # Count from 1 onward if the response has no pagination
+            if page_info['page']:
+                page = response[page_info['page']] + 1
+            else:
+                page = page + 1
+            query.update({'page': page})
+
+            if slice_end and page > slice_end:
+                # Skip requests for pages after the end of the slice
+                break
+
             response = self.__plenty_api_request(method='get',
                                                  domain=domain,
                                                  path=path,
@@ -393,13 +436,17 @@ class PlentyApi():
                 logging.error(f"subsequent {domain} API requests failed.")
                 return response
 
+            if slice_start and page < slice_start:
+                # Skip pages before the beginning of the slice
+                continue
+
             entries += response[page_info['data']]
 
             if self.cli_progress_bar:
                 if pbar:
                     pbar.update(1)
 
-        if self.cli_progress_bar:
+        if self.cli_progress_bar and page_info['last_page']:
             if pbar:
                 pbar.close()
 
@@ -474,6 +521,100 @@ class PlentyApi():
 
         return orders
 
+    def plenty_api_dump_bi_raw_file(
+        self, remote_files: Union[str, dict, list],
+        download_directory: Path = Path('.')
+    ) -> List[Union[Path, None]]:
+        """
+        Dumping BI raw data to file.
+
+        Parameter:
+
+            remote_files[str, dict, list]   -   BI files to download
+                                                determines the path to the raw
+                                                data this can be a string, a
+                                                list of strings or a dict (e.g.
+                                                the response from
+                                                `plenty_api_get_bi_raw_files`)
+        OPTIONAL
+            download_directory  [Path]      -   Local filepath to store the data
+                                                default is the current directory.
+                                                Automatically created when not
+                                                found.
+
+        Return:
+                                [list]      -   list of paths to downloaded files
+        """
+        if isinstance(remote_files, str):
+            remote_files = [remote_files]
+
+        response_list = []
+        for remote_file in remote_files:
+            remote_file_path_query = None
+            # if a response from plenty_api_get_bi_raw_files was passed, extract path
+            if isinstance(remote_file, dict) and 'path' in remote_file:
+                remote_file_path_query = Path(remote_file['path'])
+
+            if isinstance(remote_file, str):
+                remote_file_path_query = Path(remote_file)
+
+            if not remote_file_path_query:
+                logging.warning(
+                    f'validation error - invalid path query ({remote_file}) for'
+                    'BI files. Required is either a single path, a list of'
+                    'paths or a the response format of'
+                    '`plenty_api_get_bi_raw_files`.'
+                )
+                response_list.append(None)
+                continue
+            response = self.__plenty_api_request(
+                method='get', domain='bi_raw', path='/file',
+                query={'path':remote_file_path_query}
+            )
+            if isinstance(response, bytes):
+                download_directory.mkdir(parents=True, exist_ok=True)
+                file_destination = download_directory / remote_file_path_query.name
+                with open(file_destination, "wb") as requested_file:
+                        requested_file.write(response)
+                        response_list.append(file_destination)
+            else:
+                response_list.append(None)
+                logging.error('Failed to fetch {}'.format(remote_file_path_query))
+        return response_list
+
+    def plenty_api_get_bi_raw_files(self, refine: dict = None, query: dict = None) -> list:
+        """
+        Get a list of BI-Rawdata files
+
+        Parameters:
+            refine              [dict]      -   Refine arguments for the BI
+            query               [dict]      -   Query arguments e.g. page-slicing
+
+        Return:
+                        [JSON(Dict) / DataFrame] <= self.data_format
+        """
+        query = utils.sanity_check_parameter(domain='bi_raw',
+                                             query=query,
+                                             refine=refine,
+                                             additional=None)
+
+        # Prefer 100 items per page over the default of 20
+        if 'itemsPerPage' not in query:
+            query.update({'itemsPerPage':100})
+
+        bi_files = self.__repeat_get_request_for_all_records(domain='bi_raw',
+                                                             query=query)
+
+        if isinstance(bi_files, dict) and 'error' in bi_files.keys():
+            logging.error("GET BI-Rawfile list failed with:\n"
+                          f"{bi_files}")
+            return None
+
+        bi_files = utils.transform_data_type(data=bi_files,
+                                           data_format=self.data_format)
+
+        return bi_files
+
     def plenty_api_get_pending_redistribution(
         self, order_id: int = 0, sender: int = 0, receiver: int = 0,
         shipping_packages: str = ''
@@ -1176,6 +1317,16 @@ class PlentyApi():
         return utils.summarize_shipment_packages(
             response=package_responses, mode=mode)
 
+    def plenty_api_get_amazon_product_types(self):
+        """
+        Get a list of all available Amazon product types.
+
+        Return:
+                        [JSON(Dict) / DataFrame] <= self.data_format
+        """
+        return self.__plenty_api_generic_get(domain='pim',
+                                             path='/amazon-product-types')
+
 # POST REQUESTS
 
     def plenty_api_set_image_availability(self,
